//
// C++ Interface: lineparser
//
// Description: 
//
//
// Author: Eric Torstenson <torstenson@chgr.mc.vanderbilt.edu>, (C) 2006
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef ESELINEPARSER_H
#define ESELINEPARSER_H

#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include "types.h"
#include "strings.h"
#include <map>


namespace Utility {

using namespace std;


/**
 * @brief Baseclass for the ascii parser routines
 */
class AsciiParser {
public:
	AsciiParser(){}
	virtual ~AsciiParser() {}
	virtual bool ParseLine(const char *line, uint val)=0;
};


/**
  @brief Breaks the line into a vector and passes it to the functor
 */
class ArrayedTextParser : public AsciiParser {
public:
	ArrayedTextParser() {}
	virtual ~ArrayedTextParser() {} 
	bool ParseLine(const char *line, uint val);
	virtual bool Evaluate(StringArray& words, uint val)=0;
};



/**
@brief Parses ascii files and executes a functor for each line

	@author Eric Torstenson <torstenson@chgr.mc.vanderbilt.edu>
*/
class LineParser{
public:
    LineParser(char cmtChar = 0) : cmtChar(cmtChar) {}

    ~LineParser(){}

	/**
	 * @brief Opens a file and sends each line to the function pointer at fn
	 */
	uint Parse(const char* filename, AsciiParser *parser, bool keepComments = false);

	string comments;
protected:
	char cmtChar;
};

/**
  @brief Converts a file to an array of strings
 */
class FileToArray : public AsciiParser {
public:
	FileToArray();
	~FileToArray();
	
	bool ParseLine(const char *line, uint val=0);
	StringArray strings;
};



class FileToMap: public AsciiParser {
public:

	FileToMap();
	~FileToMap();

	bool ParseLine(const char *line, uint val=0);


	/**
	 * @brief Opens file and parses it, returning the number of valie lines
	 * @param filename The file to be parsed
	 * @param commentChar If the first character matches this, the line is treated as a comment
	 * @param true indicates that all comments will be stored as a single large string
	 */
	uint Parse(const char *filename, char commentChar, bool keepComments);
	/**
	 * @brief Returns the vector of lines associated with a given key. The key will remain as the first member of the line
	 * @param key The first word in the line
	 * @param lines a vector containing each string that was found
	 * @return t/f indicating that the key was found
	 * @note lines will not be changed if the return value is False. 
	 * It is important to note that the entire contents of the file will be loaded into memory.
	 */
	bool GetLines(const char *key, vector<string>& lines);

	double GetDouble(const char *key);

	int GetInteger(const char *key);

	string GetLine(const char *key);

	vector<string> GetKeys() { return keys; }

	string comments;
protected:
	map<string, vector<string> > strings;
	vector<string> keys;

};



/**
 * @brief A simple parser object that returns the number of lines encountered. 
 * Eventually, this could be expanded to get other types of information like the number of columns in certain types of lines
 */
class BasicLineCounter: public AsciiParser {
public:
	BasicLineCounter() {}
	~BasicLineCounter() {}

	bool ParseLine(const char *line, uint val=0);
};


/**
@brief Sets values in the bitvector to value if they appear in the file
*/
class ExclusionList : public AsciiParser {
public:
	ExclusionList(BitSetType *active, bool value=false) : active(active), value(value) {}
	~ExclusionList() {}
	
	bool ParseLine(const char *line, uint val=0);
protected:
	BitSetType *active;
	bool value;

};


inline
bool ExclusionList::ParseLine(const char* line, uint val /*=0*/) {
	bool success = strlen(line) > 0;

	if (success)	{
		uint idx=atoi(line);	
		(*active)[idx-1]=value;
	}
	return success;
}



inline
bool ArrayedTextParser::ParseLine(const char *rawline, uint val) {
	StringArray line;
	size_t len=strlen(rawline);
	
	if (len > 0) {
		TokenizeString(rawline, line, " \t,\r");
		return Evaluate(line, val);
	}	
	return false;
}

inline
FileToMap::FileToMap() { }

inline
FileToMap::~FileToMap() { }

inline
bool FileToMap::GetLines(const char *key, vector<string>& lines) {
	map<string, vector<string> >::iterator itr = strings.find(key);
	map<string, vector<string> >::iterator end = strings.end();

	bool isFound = false;
	if (itr != end) {
		isFound = true;
		lines=itr->second;
	}
	return isFound;
}


inline
string FileToMap::GetLine(const char *key) {
	string value="";
	vector<string> lines;

	if (GetLines(key, lines) && lines.size()>0)
		value=lines[0];
	return value;
}

inline
int FileToMap::GetInteger(const char *key) {
	return atoi(GetLine(key).c_str());
}

inline
double FileToMap::GetDouble(const char *key) {
	return atof(GetLine(key).c_str());
}




inline 
bool FileToMap::ParseLine(const char *line, uint val /* = 0*/) {
	stringstream source(line);

	bool valid = strlen(line) > 0 && line[0] != '\n';

	if (valid) {
		string key, word;

		source>>key;
	
		if (strings.find(key) == strings.end()) 
			keys.push_back(key);
		
		//We don't really care if contents is empty, 
		//we still want the key to be found in case it is alone on the line
		vector<string>& contents = strings[key];

		while (!source.eof()) {
			//On 'IX machines, sometimes that extra windows return thing will 
			//prevent us from seeing the end of line on time....gotta clean out
			//the old value
			word = "";					
			source>>word;
			if (word.length() > 0) {
				contents.push_back(word);
			}
		}
	}
	return valid;
}
		
inline
FileToArray::FileToArray() {}

inline
FileToArray::~FileToArray() {}

inline
bool FileToArray::ParseLine(const char *line, uint val) {
	strings.push_back(string(line));
	return true;
}

inline
bool BasicLineCounter::ParseLine(const char *line, uint val) {
	if (strlen(line) > 0 && line[0] != '/' && line[0] != '#')
		return true;
	else
		return false;
}

inline
uint FileToMap::Parse(const char *filename, char ignoreChar, bool keepComments) {
	LineParser lp(ignoreChar);
	keys.clear();
	uint linesKept = lp.Parse(filename, this, keepComments);
	if (keepComments)
		comments = lp.comments;
	return linesKept;
}


}

#endif
